'''Train LSTM RNNs on the AIDR tweet classification task.

GPU command:
    THEANO_FLAGS=mode=FAST_RUN,device=gpu,floatX=float32 python lstm_rnns_aidr.py

Output after 4 epochs on CPU: ~0.8146
Time per epoch on CPU (Core i7): ~150s.
'''

# keras related
from __future__ import print_function
import numpy as np
np.random.seed(2000)  # for reproducibility

from keras.preprocessing import sequence
from keras.utils.np_utils import accuracy, to_categorical
from keras.models import model_from_json

from keras.models import Sequential
from keras.layers.core import Dense, Dropout, Activation
from keras.optimizers import SGD, Adam, RMSprop


from utilities import aidr_feat
from keras.callbacks import EarlyStopping, ModelCheckpoint
from sklearn import metrics

#other utilities 
import optparse
import logging
import sys


def build_MLP(model_type, max_features, nb_layers=1, hidden_size=128, nb_classes = 2, dropout_ratio=0.5):

	''' build MLP '''

	model = Sequential()

	if nb_layers == 0: #no hidden layer
		print('Doing LR classification with class #', nb_classes)
		model.add(Dense(nb_classes, input_shape=(max_features, ))) 
		model.add(Activation('softmax'))

	elif nb_layers > 0: # add one hidden layer
		model.add(Dense(hidden_size, input_shape=(max_features, ))) 
		model.add(Activation('relu'))	
		model.add(Dropout(dropout_ratio))
	
	elif nb_layers > 1: # add another hidden layer
		model.add(Dense(hidden_size))
		model.add(Activation('relu'))
		model.add(Dropout(dropout_ratio))
	else:
		print("layer nb not supported..", nb_layers)
		exit(1)


	# We project onto a single unit output layer, and squash it with a sigmoid:
	if nb_classes == 2 and nb_layers != 0:
		print('Doing binary classification...')
		model.add(Dense(1))
		model.add(Activation('sigmoid'))

	elif nb_classes > 2 and nb_layers != 0:
		print('Doing classification with class #', nb_classes)
		model.add(Dense(nb_classes))
		model.add(Activation('softmax'))


	return model	


if __name__ == '__main__':


	logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s', datefmt='%Y-%m-%d %H:%M:%S', level=logging.DEBUG)

	# parse user input
	parser = optparse.OptionParser("%prog [options]")

	#file related options
	parser.add_option("-g", "--log-file",          dest="log_file", help="log file [default: %default]")
	parser.add_option("-d", "--data-dir",          dest="data_dir", help="directory containing train, test and dev file [default: %default]")
	parser.add_option("-D", "--data-spec",         dest="data_spec", help="specification for training data (in, out, in_out) [default: %default]")
	parser.add_option("-p", "--model-dir",          dest="model_dir", help="directory to save the best models [default: %default]")

	# adp related
	parser.add_option("-m", "--model-path",         dest="model_path", help="path to the  model [default: %default]")
	parser.add_option("-O", "--od-data-dir",        dest="od_data_dir", help="od data directory containing train, test and dev file [default: %default]")

	# network related
	parser.add_option("-s", "--hidden-size",       dest="hidden_size",   type="int", help="hidden layer size [default: %default]") # size of the hidden layer
	parser.add_option("-o", "--dropout_ratio",     dest="dropout_ratio", type="float", help="ratio of cells to drop out [default: %default]")

	#learning related
	parser.add_option("-a", "--learning-algorithm", dest="learn_alg", help="optimization algorithm (adam, sgd, adagrad, rmsprop, adadelta) [default: %default]")
	parser.add_option("-b", "--minibatch-size",     dest="minibatch_size", type="int", help="minibatch size [default: %default]")
	parser.add_option("-l", "--loss",               dest="loss", help="loss type (hinge, squared_hinge, binary_crossentropy) [default: %default]")
	parser.add_option("-n", "--epochs",             dest="epochs", type="int", help="nb of epochs [default: %default]")
	parser.add_option("-C", "--map-class",          dest="map_class", type="int", help="map classes to five labels [default: %default]")
	parser.add_option("-H", "--nb-layers",          dest="nb_layers", type="int", help="nb of hidden layers [default: %default]")
	parser.add_option("-F", "--add-feature",        dest="add_feat",  type="int", help="add extra features [default: %default]")

	parser.set_defaults(
    	data_dir        = None
    	,data_spec       = "in"
    	,model_dir       = "./saved_models/"
    	,model_path     = None
    	,od_data_dir    = None
	    ,log_file       = "log"
	   	,learn_alg      = "adam" # sgd, adagrad, rmsprop, adadelta, adam (default)
	   	,loss           = "binary_crossentropy" # hinge, squared_hinge, binary_crossentropy (default)
	    ,minibatch_size = 32
    	,dropout_ratio  = 0.0
    	,epochs         = 25
    	,hidden_size    = 128
    	,nb_layers      = 0
    	,model_type     = 'mlp' #bidirectional, unidirectional (default)
    	,add_feat       = 0
    	,map_class      = 0
	)

	options,args = parser.parse_args(sys.argv)

	# load out-domain data
	if options.od_data_dir:
		print('Step 1: Loading od or cat data...', options.od_data_dir)
		(X_train_od, y_train_od), (X_test_od, y_test_od), (X_dev, y_dev), label_id = aidr_feat.load_data_unig(path=options.od_data_dir,\
															map_labels_to_five_class=options.map_class, add_feat=options.add_feat) # for unigram feature file

		y_test_od   = np.array(y_test_od)
		y_train_od  = np.array(y_train_od)
		y_dev       = np.array(y_dev) # dev will be out-domain dev
		
		print('Done ...')
	else:
		print ("Error!!! provide a directory of out-domain/cat data...")
		exit(1)

	nb_classes = len(label_id)
	max_features = X_train_od.shape[1]
	loss       = 'categorical_crossentropy'
	optimizer  = 'adadelta'

	# load in-domain model
	if options.model_path:
		try:
			print('Step 2: Loading and compiling provided model...')
			in_model = model_from_json(open(options.model_path + ".json").read())
			in_model.load_weights(options.model_path)
			in_model.compile(optimizer=optimizer, loss=loss,  metrics=['accuracy'])
			print('Done ...')

		except IOError as e:
			print (options.model_path)
			print ("I/O error({0}): {1}".format(e.errno, e.strerror)) 	

	else:
		print ("Error!!! provide a model to use for scoring out-domain data...")
		exit(1)


	print('Step 3: Applying in-model: ', options.model_path, ' to od-data to score:', options.od_data_dir)

	# perform some sanity check before applying 
	if in_model.layers[0].input_shape[1] != max_features:
		print ("Error!!! the input feature dims aren't same...")
		exit (1)

	if in_model.layers[-2].output_dim != nb_classes:
		print ("Error!!! output dims don't match ...")
		exit (1)

	#validate the results by applying first on the test data
	y_pred_od      = in_model.predict(X_test_od) # preds over K-classes
	y_pred_od_cl   = np.argmax(y_pred_od, axis=1)
	acc = metrics.accuracy_score(y_test_od, y_pred_od_cl)
	print("Test ID set raw Accuracy:", acc)

	y_pred_od      = in_model.predict(X_train_od) # preds over K-classes
	y_pred_od_cl   = np.argmax(y_pred_od, axis=1)
	acc = metrics.accuracy_score(y_train_od, y_pred_od_cl)
	print("Train OD set raw Accuracy:", acc)

	y_train_od_mod  = to_categorical(y_train_od,  nb_classes) # 1-hot representation 
	y_prob_od       = y_pred_od * y_train_od_mod # take the relevant score only in 1-hot 

	# load in-domain data
	if options.data_dir:
		print('Step 4: Loading id data...', options.data_dir)
		(X_train_id, y_train_id), (X_test, y_test), (X_dev_id, y_dev_id), label_id = aidr_feat.load_data_unig(path=options.data_dir,\
															map_labels_to_five_class=options.map_class, add_feat=options.add_feat) # for unigram feature file
		y_train_id  = np.array(y_train_id)
		y_test      = np.array(y_test) #test will be in-domain test
		y_train_id_mod = to_categorical(y_train_id,  nb_classes) # 1-hot representation 

		print('Done ...')
	else:
		print ("Error!!! provide a directory of in-domain data...")
		exit(1)

	# concatenate data
	X_train = np.concatenate((X_train_id, X_train_od), axis=0)
	y_train = np.concatenate((y_train_id, y_train_od), axis=0)
	y_prob  = np.concatenate((y_train_id_mod, y_prob_od)) # in-dom weight will be 1 and out domin is prob

	y_train_mod = to_categorical(y_train,  nb_classes)
	y_dev_mod   = to_categorical(y_dev, nb_classes)
	y_test_mod  = to_categorical(y_test, nb_classes)

	val_prob    = np.zeros_like(y_dev_mod, dtype=np.float) # requires for training

	assert y_prob.shape == y_train_mod.shape

	#build model...
	print('............................')
	print(len(X_train), 'train tweets')
	print(len(X_test),  'test  tweets')
	print(len(X_dev),   'dev   tweets')
	print(max_features, 'features')
	print(nb_classes, 'different classes')
	print('............................')

	print ('Step 5: Building adapted model ')

	model = build_MLP(options.model_type, max_features, nb_layers=options.nb_layers, hidden_size=options.hidden_size,\
						nb_classes=nb_classes, dropout_ratio=options.dropout_ratio)

	model.compile(optimizer=optimizer, loss='my_categorical_crossentropy',  metrics=['accuracy'], reg_flag=1) 
#	model.compile(optimizer=optimizer, loss='categorical_crossentropy',  metrics=['accuracy']) 

#		# dummy probs for reg
#		y_prob      = np.random.uniform(low=0.1, high=1.0, size=y_train_mod.shape) * y_train_mod

	model_name = options.model_dir + "MLP-ADP-" + "-layers-" + str(options.nb_layers) + "-hidden-" + str (options.hidden_size) + "-"\
					+ optimizer + "-" + loss + "-" + str (options.minibatch_size) + "-" + str(options.dropout_ratio) + "-"\
					+ str (max_features) + "-" + ".model.cl." + str(nb_classes) + ".dom." + options.data_spec 

	earlystopper = EarlyStopping(monitor='val_acc', patience=5, verbose=1)
	checkpointer = ModelCheckpoint(filepath=model_name, monitor='val_acc', verbose=1, save_best_only=True)

	print ('Training and validating ....')
	model.fit(X_train, y_train_mod, probs=y_prob, val_probs=val_prob, batch_size=options.minibatch_size, nb_epoch=options.epochs,
			validation_data=(X_dev, y_dev_mod), verbose=2, callbacks=[earlystopper, checkpointer])

#	model.fit(X_train, y_train_mod, batch_size=options.minibatch_size, nb_epoch=options.epochs,
#			validation_data=(X_dev, y_dev_mod), verbose=2, callbacks=[earlystopper, checkpointer])


	print("Test model ...")
	print ("Loading ...", model_name)
	model.load_weights(model_name)

	y_pred = model.predict_classes(X_test)
	
	acc2 = metrics.accuracy_score(y_test, y_pred)
	print("Raw Accuracy:", acc2)

	#get label ids in sorted
	class_labels = sorted(label_id, key=label_id.get)

#	class_labels = map(str, sorted(label_id, key=label_id.get))
	class_ids    = sorted(label_id.values())

	print (metrics.classification_report(y_test, y_pred, labels=class_ids, target_names=class_labels) )

	print ("Confusion Matrix:\n", metrics.confusion_matrix(y_test, y_pred) )

	mic_p, mic_r, mic_f, sup = metrics.precision_recall_fscore_support(y_test, y_pred, average='micro')
	mac_p, mac_r, mac_f, sup = metrics.precision_recall_fscore_support(y_test, y_pred, average='macro')

	print (" micro pre: " + str (mic_p) + " rec: " + str (mic_r) + " f-score: " + str (mic_f))
	print (" macro pre: " + str (mac_p) + " rec: " + str (mac_r) + " f-score: " + str (mac_f))

	# save the architecture finally in json format
	json_string = model.to_json()
	open(model_name + ".json", 'w').write(json_string)


